import os
import subprocess
import sys
import platform
import logging

# Add and format time stamp in logging messages
logging.basicConfig(
    format="%(asctime)s %(levelname)s %(message)s",
    level=logging.INFO,
    datefmt="%c",
)
# Mute numexpr threads info
logging.getLogger("numexpr").setLevel(logging.WARNING)

# Constants
from .constants import MUSCLE_GITHUB_LINK

# Get absolute package path
PACKAGE_PATH = os.path.abspath(os.path.dirname(__file__))
# Path to muscle binary (only exists after 'compile_muscle' was executed)
MUSCLE_PATH = os.path.join(
    PACKAGE_PATH, f"bins/compiled/muscle/src/{platform.system()}/muscle"
)


def compile_muscle():
    """
    Compiles MUSCLE from source.
    Currently only supports Linux and Darwin.
    """

    if platform.system() != "Linux" and platform.system() != "Darwin":
        raise OSError(
            f"Muscle compiler currently only supports Linux and Darwin, not {platform.system()}.\n"
        )

    logging.info("Compiling MUSCLE binary from source... ")

    # Record current working directory
    cwd = os.getcwd()

    # Change path to package path
    os.chdir(PACKAGE_PATH)
    # Create folders 'bins/compiled/' inside gget package
    os.makedirs("bins/compiled/", exist_ok=True)
    # Change path to PACKAGE_PATH/bins/compiled/
    os.chdir("bins/compiled/")

    # Clone MUSCLE repo into PACKAGE_PATH/bins/compiled/
    command1 = "git clone " + MUSCLE_GITHUB_LINK + " -q"

    with subprocess.Popen(command1, shell=True, stderr=subprocess.PIPE) as process_1:
        stderr_1 = process_1.stderr.read().decode("utf-8")
        # Log the standard error if it is not empty
        if stderr_1:
            sys.stderr.write(stderr_1)
    # Exit system if the subprocess returned with an error
    if process_1.wait() != 0:
        sys.exit(f"'{command1}' command returned with error {process_1.wait()}.")

    # Change path to PACKAGE_PATH/bins/compiled/muscle/src/
    os.chdir("muscle/src/")

    # Run make command
    if platform.system() == "Linux":
        logging.warning(
            "Compiling MUSCLE requires that g++, make, sed and git are installed."
        )
    if platform.system() == "Darwin":
        logging.warning(
            "Compiling MUSCLE requires that gcc v11, make, sed and git are installed."
        )
        logging.warning(
            "Please run 'brew install gcc' to install gcc v11 if the compile fails."
        )

    command2 = "make -s"

    with subprocess.Popen(command2, shell=True, stderr=subprocess.PIPE) as process_2:
        stderr_2 = process_2.stderr.read().decode("utf-8")
        # Log the standard error if it is not empty
        if stderr_2:
            sys.stderr.write(stderr_2)
    # Exit system if the subprocess returned with an error
    if process_2.wait() != 0:
        sys.exit(f"'{command2}' command returned with error {process_2.wait()}.")

    logging.info("MUSCLE compiled.")

    # Change path back to cwd
    os.chdir(cwd)



import logging

# Add and format time stamp in logging messages
logging.basicConfig(
    format="%(asctime)s %(levelname)s %(message)s",
    level=logging.INFO,
    datefmt="%c",
)
# Mute numexpr threads info
logging.getLogger("numexpr").setLevel(logging.WARNING)

import json as json_package
from json.decoder import JSONDecodeError
import pandas as pd
from urllib.request import urlopen


def blat(sequence, seqtype="default", assembly="human", json=False, save=False):
    """
    BLAT a nucleotide or amino acid sequence against any BLAT UCSC assembly.

    Args:
     - sequence       Sequence (str) or path to fasta file containing one sequence.
     - seqtype        'DNA', 'protein', 'translated%20RNA', or 'translated%20DNA'.
                      Default: 'DNA' for nucleotide sequences; 'protein' for amino acid sequences.
     - assembly       'human' (hg38) (default), 'mouse' (mm39), 'zebrafinch' (taeGut2),
                      or any of the species assemblies available at https://genome.ucsc.edu/cgi-bin/hgBlat
                      (use short assembly name as listed after the "/").
     - json           If True, returns results in json format instead of data frame. Default: False.
     - save           If True, the data frame is saved as a csv in the current directory (default: False).

    Returns a data frame with the BLAT results.
    """

    ## Clean up sequence
    # If the path to a fasta file was provided instead of a nucleotide sequence,
    # read the file and extract the first sequence
    if "." in sequence:
        if ".txt" in sequence:
            # Read the text file
            titles = []
            seqs = []
            with open(sequence) as text_file:
                for i, line in enumerate(text_file):
                    # Recognize a title line by the '>' character
                    if line[0] == ">":
                        # Append title line to titles list
                        titles.append(line.strip())
                    else:
                        seqs.append(line.strip())

        elif ".fa" in sequence:
            # Read the FASTA
            titles = []
            seqs = []
            with open(sequence) as fasta_file:
                for i, line in enumerate(fasta_file):
                    # Each second line will be a title line
                    if i % 2 == 0:
                        if line[0] != ">":
                            raise ValueError(
                                "Expected FASTA to start with a '>' character. "
                            )
                        else:
                            # Append title line to titles list
                            titles.append(line.strip())
                    else:
                        if line[0] == ">":
                            raise ValueError(
                                "FASTA contains two lines starting with '>' in a row -> missing sequence line. "
                            )
                        # Append sequences line to seqs list
                        else:
                            seqs.append(line.strip())
        else:
            raise ValueError(
                "File format not recognized. gget BLAT currently only supports '.txt' or '.fa' files. "
            )

        # Set the first sequence from the fasta file as 'sequence'
        sequence = seqs[0]
        if len(seqs) > 1:
            logging.info(
                "File contains more than one sequence. Only the first sequence will be submitted to BLAT."
            )

    # Shorten sequence to length limit if necessary
    if len(sequence) > 8000:
        logging.info(
            "Length of sequence is > 8000. Only the fist 8000 characters will be submitted to BLAT."
        )
        sequence = sequence[:8000]

    ## Set seqtype
    # Valid seqtype options
    seqtypes = ["DNA", "protein", "translated%20RNA", "translated%20DNA"]

    # If user does not specify the seqtype,
    # check if a nucleotide or amino acid sequence was passed
    if seqtype == "default":
        # Set of all possible nucleotides and amino acids
        nucleotides = set("ATGC")
        amino_acids = set("ARNDCQEGHILKMFPSTWYVBZ")

        # If sequence is a nucleotide sequence, set seqtype to DNA
        if set(sequence) <= nucleotides:
            seqtype = "DNA"
            logging.info(
                f"Sequence recognized as nucleotide sequence. 'seqtype' will be set as {seqtype}."
            )

        # If sequence is an amino acid sequence, set seqtype to protein
        elif set(sequence) <= amino_acids:
            seqtype = "protein"
            logging.info(
                f"Sequence recognized as amino acid sequence. 'seqtype' will be set as {seqtype}."
            )

        else:
            raise ValueError(
                f"""
                Sequence not automatically recognized as a nucleotide or amino acid sequence.
                Please specify 'seqtype'.
                Seqtype options: {', '.join(seqtypes)} 
                """
            )

    else:
        # Check if the user specified seqtype is valid
        if seqtype not in seqtypes:
            raise ValueError(
                f"Seqtype specified is {seqtype}. Expected one of {', '.join(seqtypes)}"
            )

    ## Set assembly
    # Note: If assembly not found, defaults to hg38
    if assembly == "human" or assembly == "homo_sapiens":
        database = "hg38"
    elif assembly == "mouse" or assembly == "mus_musculus":
        database = "mm39"
    elif assembly == "zebrafinch" or assembly == "taeniopygia_guttata":
        database = "taeGut2"
    else:
        database = assembly

    # Define server URL
    url = f"https://genome.ucsc.edu/cgi-bin/hgBlat?userSeq={sequence}&type={seqtype}&db={database}&output=json"

    # Submit URL request
    r = urlopen(url)
    if r.status != 200:
        raise RuntimeError(
            f"HTTP response status code {r.status}. "
            "Please double-check arguments and try again.\n"
        )

    try:
        # Read json results into a dictionary
        results = json_package.load(r)
    except JSONDecodeError:
        logging.error(
            f"""
            BLAT of seqtype '{seqtype}' using assembly '{database}' was unsuccesful. 
            Possible causes: 
            - Sequence possibly too short (required minimum: 20 characters). 
            - Assembly possibly invalid. All available species with their respective assemblies are listed at https://genome.ucsc.edu/cgi-bin/hgBlat.
            """
        )
        return

    if len(results["blat"]) == 0:
        logging.info(
            f"No {seqtype} BLAT matches were found for this sequence in genome {results['genome']}."
        )
        return

    # Let user know if assembly was not found
    # If this is the case, BLAT automatically defaults to human (hg38)
    if results["genome"] != database:
        logging.warning(
            f"Assembly {database} not recognized. Defaulted to {results['genome']} instead."
        )

    ## Build data frame to resemble BLAT web search results
    # Define dataframe from dictionary
    df_dict = {}

    for field in results["fields"]:
        df_dict.update({field: []})

    for blat_result_list in results["blat"]:
        for field, (i, result) in zip(results["fields"], enumerate(blat_result_list)):
            df_dict[field].append(result)

    df = pd.DataFrame(df_dict)

    # Calculate % aligned sequence of submitted sequence
    aligned_size = df["qEnd"] - df["qStart"]
    df["%_aligned"] = round((100 / df["qSize"]) * aligned_size, 2)
    # Calculate % matched sequence of aligned sequence
    df["%_matched"] = round((100 / aligned_size) * df["matches"], 2)
    # Add genome column
    df["genome"] = results["genome"]

    # Adjust sequence start to match website
    df["qStart"] = df["qStart"] + 1
    df["tStart"] = df["tStart"] + 1

    # Rename columns
    columns_dict = {
        "misMatches": "mismatches",
        "qName": "query",
        "qSize": "query_size",
        "qStart": "aligned_start",
        "qEnd": "aligned_end",
        "tName": "chromosome",
        "tStart": "start",
        "tEnd": "end",
    }

    df = df.rename(columns=columns_dict)

    # Change columns order (this also drops all unmentioned columns)
    df = df.reindex(
        columns=[
            "genome",
            "query_size",
            "aligned_start",
            "aligned_end",
            "matches",
            "mismatches",
            "%_aligned",
            "%_matched",
            "chromosome",
            "strand",
            "start",
            "end",
        ]
    )

    if json:
        results_dict = json_package.loads(df.to_json(orient="records"))
        if save:
            with open("gget_blat_results.json", "w", encoding="utf-8") as f:
                json_package.dump(results_dict, f, ensure_ascii=False, indent=4)

        return results_dict

    else:
        if save:
            df.to_csv("gget_blat_results.csv", index=False)

        return df



import os
import shutil
import sys
import subprocess
import platform
import uuid
from platform import python_version

import logging

logging.basicConfig(
    format="%(asctime)s %(levelname)s %(message)s",
    level=logging.INFO,
    datefmt="%c",
)
# Mute numexpr threads info
logging.getLogger("numexpr").setLevel(logging.WARNING)

from .compile import PACKAGE_PATH

## Variables for alphafold module
ALPHAFOLD_GIT_REPO = "https://github.com/deepmind/alphafold"
PDBFIXER_GIT_REPO = "https://github.com/openmm/pdbfixer.git"
# Unique ID to name temporary jackhmmer folder
UUID = "fcb45c67-8b27-4156-bbd8-9d11512babf2"
# Path to temporary mounted disk (global)
TMP_DISK = ""
# Model parameters
PARAMS_URL = (
    "https://storage.googleapis.com/alphafold/alphafold_params_colab_2022-03-02.tar"
)
PARAMS_DIR = os.path.join(PACKAGE_PATH, "bins/alphafold/")
PARAMS_PATH = os.path.join(PARAMS_DIR, "params_temp.tar")


def setup(module):
    """
    Function to install third-party dependencies for a specified gget module.

    Args:
    module      - (str) gget module for which dependencies should be installed, e.g. "alphafold"
    """
    supported_modules = ["alphafold"]
    if module not in supported_modules:
        raise ValueError(
            f"'module' argument specified as {module}. Expected one of: {', '.join(supported_modules)}"
        )

    if module == "alphafold":
        ## Make sure package paths are appended so openmm can be imported
        site_packages_path = os.__file__.split("os.py")[0] + "site-packages"
        if site_packages_path not in sys.path:
            sys.path.append(site_packages_path)

        site_packages_path_python = (
            "/".join(str(os.__file__.split("os.py")[0]).split("/")[:-2])
            + f"/python{'.'.join(python_version().split('.')[:2])}/site-packages"
        )
        if site_packages_path_python not in sys.path:
            sys.path.append(site_packages_path_python)

        site_packages_path_python37 = (
            "/".join(str(os.__file__.split("os.py")[0]).split("/")[:-2])
            + f"/python3.7/site-packages"
        )
        if site_packages_path_python37 not in sys.path:
            sys.path.append(site_packages_path_python37)

        conda_python_path = os.path.expanduser(
            f"~/opt/conda/lib/python{'.'.join(python_version().split('.')[:2])}/site-packages"
        )
        if conda_python_path not in sys.path:
            sys.path.append(conda_python_path)

        conda_python37_path = os.path.expanduser(
            f"~/opt/conda/lib/python3.7/site-packages"
        )
        if conda_python37_path not in sys.path:
            sys.path.append(conda_python37_path)

        # Global location of temporary disk
        global TMP_DISK

        ## Ask user to install openmm if not already installed
        try:
            import simtk.openmm as openmm

            # Silence openmm logger
            logging.getLogger("openmm").setLevel(logging.WARNING)

            # Check if correct version was installed
            if openmm.__version__ != "7.5.1":
                raise ImportError()

            logging.info(f"openmm v{openmm.__version__} already installed.")

        except ImportError:
            logging.error(
                """
        Please install third-party dependency openmm v7.5.1 by running the following command from the command line:
        'conda install -c conda-forge openmm=7.5.1'
        """
            )
            return

        ## Install Alphafold if not already installed
        logging.info("Installing AlphaFold from source (requires pip and git).")
        # Install AlphaFold and apply OpenMM patch.
        # command = f"""
        #     git clone {ALPHAFOLD_GIT_REPO} alphafold \
        #     && sed -i '' 's/\/tmp\/ramdisk/~\/tmp\/jackhmmer\/{UUID}/g' ./alphafold/alphafold/data/tools/jackhmmer.py \
        #     && pip install ./alphafold \
        #     && pushd {os.__file__.split('os.py')[0] + 'site-packages/'} \
        #     && patch -p0 < /content/alphafold/docker/openmm.patch \
        #     && popd \
        #     && rm -rf alphafold
        #     """

        ## Install AlphaFold and change jackhmmer directory where database chunks are saved in
        # Define AlphaFold folder name and location
        alphafold_folder = os.path.join(
            PACKAGE_PATH, "tmp_alphafold_" + str(uuid.uuid4())
        )
        # command = """
        #     git clone -q {} {} \
        #     && sed -i '' 's/\/tmp\/ramdisk/{}/g' {}/alphafold/data/tools/jackhmmer.py \
        #     && sed -i '' '/{}/d' {}/alphafold/data/tools/jackhmmer.py \
        #     && pip install -q {} \
        #     """.format(
        #     ALPHAFOLD_GIT_REPO,
        #     alphafold_folder,
        #     os.path.expanduser(f"~/tmp/jackhmmer/{UUID}").replace(
        #         "/", "\/"
        #     ),  # Replace directory where jackhmmer database chunks will be saved
        #     alphafold_folder,
        #     "logging.info",  # Delete all info loggers
        #     alphafold_folder,
        #     alphafold_folder,
        # )

        # Clone AlphaFold github repo
        # Replace directory where jackhmmer database chunks will be saved
        # Insert "logging.set_verbosity(logging.WARNING)" to mute all info loggers
        # Pip install AlphaFold from local directory
        if platform.system() == "Darwin":
            command = """
                git clone -q {} {} \
                && sed -i '' 's/\/tmp\/ramdisk/{}/g' {}/alphafold/data/tools/jackhmmer.py \
                && sed -i '' 's/from absl import logging/from absl import logging\\\nlogging.set_verbosity(logging.WARNING)/g' {}/alphafold/data/tools/jackhmmer.py \
                && pip install -q {} \
                """.format(
                ALPHAFOLD_GIT_REPO,
                alphafold_folder,
                os.path.expanduser(f"~/tmp/jackhmmer/{UUID}").replace(
                    "/", "\/"
                ),  # Replace directory where jackhmmer database chunks will be saved
                alphafold_folder,
                alphafold_folder,
                alphafold_folder,
            )
        else:
            command = """
                git clone -q {} {} \
                && sed -i 's/\/tmp\/ramdisk/{}/g' {}/alphafold/data/tools/jackhmmer.py \
                && sed -i 's/from absl import logging/from absl import logging\\\nlogging.set_verbosity(logging.WARNING)/g' {}/alphafold/data/tools/jackhmmer.py \
                && pip install -q {} \
                """.format(
                ALPHAFOLD_GIT_REPO,
                alphafold_folder,
                os.path.expanduser(f"~/tmp/jackhmmer/{UUID}").replace(
                    "/", "\/"
                ),  # Replace directory where jackhmmer database chunks will be saved
                alphafold_folder,
                alphafold_folder,
                alphafold_folder,
            )

        with subprocess.Popen(command, shell=True, stderr=subprocess.PIPE) as process:
            stderr = process.stderr.read().decode("utf-8")
        # Exit system if the subprocess returned with an error
        if process.wait() != 0:
            if stderr:
                # Log the standard error if it is not empty
                sys.stderr.write(stderr)
            logging.error("AlphaFold installation failed.")
            return

        # Remove cloned directory
        shutil.rmtree(alphafold_folder)

        try:
            import alphafold as AlphaFold

            logging.info(f"AlphaFold installed succesfully.")
        except ImportError:
            logging.error("AlphaFold installation failed.")
            return

        ## Append AlphaFold to path
        alphafold_path = os.path.abspath(os.path.dirname(AlphaFold.__file__))
        if alphafold_path not in sys.path:
            sys.path.append(alphafold_path)

        ## Install pdbfixer v1.7 (compadible with openmm v7.5.1)
        logging.info("Installing pdbfixer from source (requires pip and git).")

        pdbfixer_folder = os.path.join(
            PACKAGE_PATH, "tmp_pdbfixer_" + str(uuid.uuid4())
        )
        command = f"""
            git clone -q --branch v1.7 {PDBFIXER_GIT_REPO} {pdbfixer_folder} \
            && pip install -q {pdbfixer_folder} \
            """

        with subprocess.Popen(command, shell=True, stderr=subprocess.PIPE) as process:
            stderr = process.stderr.read().decode("utf-8")
        # Exit system if the subprocess returned with an error
        if process.wait() != 0:
            if stderr:
                # Log the standard error if it is not empty
                sys.stderr.write(stderr)
            logging.error("pdbfixer installation failed.")
            return

        # Remove cloned directory
        shutil.rmtree(pdbfixer_folder)

        # Check if pdbfixer was installed successfully
        command = "pip list | grep pdbfixer"
        process = subprocess.Popen(command, shell=True, stdout=subprocess.PIPE)
        pdb_out, err = process.communicate()

        if pdb_out.decode() != "":
            logging.info(f"pdbfixer installed succesfully.")
        else:
            logging.error("pdbfixer installation failed.")
            return

        # ## Create a temporary file system (TMPFS) to store a database chunk to make Jackhmmer run fast.
        # # TMPFS uses local memory for file system reads and writes, which is typically much faster than reads and writes in a UFS file system.
        # logging.info("Creating temporary file system (TMPFS) to store a database chunk and make Jackhmmer run faster.")
        # if platform.system() == "Linux":
        #   command = f"mkdir -m 777 -p /tmp/ramdisk && mount -t tmpfs -o size=9G ramdisk /tmp/ramdisk"

        #   with subprocess.Popen(command, shell=True, stderr=subprocess.PIPE) as process:
        #       stderr = process.stderr.read().decode("utf-8")

        #   # Exit system if the subprocess returned with an error
        #   if process.wait() != 0:
        #       # if stderr:
        #       # # Log the standard error if it is not empty
        #       # sys.stderr.write(stderr)
        #       logging.warning("Creating TMPFS failed. Jackhmmer will run slower.")

        # elif platform.system() == "Darwin":
        #   # Attach disk with 9GB
        #   command1 = "hdiutil attach -nomount ram://18432000"
        #   process = subprocess.Popen(command1, shell=True, stdout=subprocess.PIPE)
        #   hdi_out, err = process.communicate()

        #   # Record number of new disk
        #   TMP_DISK = hdi_out.decode("utf-8").strip()
        #   DISK_NUMBER = f"$({TMP_DISK} | tr -dc '0-9')"

        #   # Set up TMPFS
        #   command2 = f"newfs_hfs -v tmp /dev/rdisk{DISK_NUMBER}"
        #   command3 = f"diskutil eraseVolume HFS+ /tmp/ramdisk {TMP_DISK}"

        #   command = command2 + " && " + command3

        #   with subprocess.Popen(command, shell=True, stderr=subprocess.PIPE) as process:
        #     stderr = process.stderr.read().decode("utf-8")

        #   # Exit system if the subprocess returned with an error
        #   if process.wait() != 0:
        #       # if stderr:
        #       # # Log the standard error if it is not empty
        #       # sys.stderr.write(stderr)
        #       logging.warning("Creating TMPFS failed. Jackhmmer will run slower.")

        # else:
        # # Create folder to save temporary jackhmmer database chunks in
        # !!! Delete the duplicate line from get_msa function in gget_alphafold.py if this is uncommented
        # os.makedirs(f"~/tmp/jackhmmer/{UUID}", exist_ok=True)

        ## Download model parameters
        # Download parameters if the params directory is empty
        if not os.path.exists(os.path.join(PARAMS_DIR, "params/")):
            # Create folder to save parameter files
            os.makedirs(os.path.join(PARAMS_DIR, "params/"), exist_ok=True)

        if len(os.listdir(os.path.join(PARAMS_DIR, "params/"))) < 12:
            logging.info(
                "Downloading AlphaFold model parameters (requires 4.1 GB of storage). This might take a few minutes."
            )
            if platform.system() == "Windows":
                # The double-quotation marks allow white spaces in the path, but this does not work for Windows
                command = f"""
                    curl -# -o {PARAMS_PATH} {PARAMS_URL} \
                    && tar --extract --file={PARAMS_PATH} --directory={PARAMS_DIR+'params/'} --preserve-permissions \
                    && rm {PARAMS_PATH}
                    """
            else:
                command = f"""
                    curl -# -o '{PARAMS_PATH}' '{PARAMS_URL}' \
                    && tar --extract --file='{PARAMS_PATH}' --directory='{PARAMS_DIR+'params/'}' --preserve-permissions \
                    && rm '{PARAMS_PATH}'
                    """

            with subprocess.Popen(
                command, shell=True, stderr=subprocess.PIPE
            ) as process:
                stderr = process.stderr.read().decode("utf-8")
                # Log the standard error if it is not empty
                if stderr:
                    sys.stderr.write(stderr)
            # Exit system if the subprocess returned with an error
            if process.wait() != 0:
                return
            else:
                logging.info("Model parameter download complete.")
        else:
            logging.info("AlphaFold model parameters already downloaded.")



import numpy as np
import json as json_package
import mysql.connector as sql
import time
import logging

# Add and format time stamp in logging messages
logging.basicConfig(
    format="%(asctime)s %(levelname)s %(message)s",
    level=logging.INFO,
    datefmt="%c",
)
# Mute numexpr threads info
logging.getLogger("numexpr").setLevel(logging.WARNING)

import warnings

warnings.simplefilter(action="ignore", category=UserWarning)
import pandas as pd


# Custom functions
from .utils import gget_species_options, find_latest_ens_rel, wrap_cols_func


def search(
    searchwords,
    species,
    id_type="gene",
    seqtype=None,
    andor="or",
    limit=None,
    wrap_text=False,
    json=False,
    save=False,
):
    """
    Function to query Ensembl for genes based on species and free form search terms.
    Automatically fetches results from latest Ensembl release, unless user specifies database (see 'species' argument).

    Args:
    - searchwords     Free form search words (not case-sensitive) as a string or list of strings
                      (e.g.searchwords = ["GABA", "gamma-aminobutyric"]).
    - species         Species can be passed in the format "genus_species", e.g. "homo_sapiens".
                      To pass a specific database, enter the name of the core database, e.g. 'mus_musculus_dba2j_core_105_1'.
                      All availabale species databases can be found here: http://ftp.ensembl.org/pub/release-106/mysql/
    - id_type         "gene" (default) or "transcript"
                      Defines whether genes or transcripts matching the searchwords are returned.
    - andor           "or" (default) or "and"
                      "or": Returns all genes that INCLUDE AT LEAST ONE of the searchwords in their name/description.
                      "and": Returns only genes that INCLUDE ALL of the searchwords in their name/description.
    - limit           (int) Limit the number of search results returned (default: None).
    - wrap_text       If True, displays data frame with wrapped text for easy reading. Default: False.
    - json            If True, returns results in json format instead of data frame. Default: False.
    - save            If True, the data frame is saved as a csv in the current directory (default: False).

    Returns a data frame with the query results.

    Deprecated arguments: 'seqtype' (renamed to id_type)
    """
    # Handle deprecated arguments
    if seqtype:
        logging.error(
            "'seqtype' argument deprecated! Please use argument 'id_type' instead."
        )
        return

    start_time = time.time()

    # Find latest Ensembl release
    ens_rel = find_latest_ens_rel()

    ## Check validity or arguments
    # Check if id_type is valid
    id_types = ["gene", "transcript"]
    id_type = id_type.lower()
    if id_type not in id_types:
        raise ValueError(
            f"ID type (id_type) specified is '{id_type}'. Expected one of: {', '.join(id_types)}"
        )

    # Check if 'andor' arg is valid
    andors = ["and", "or"]
    andor = andor.lower()
    if andor not in andors:
        raise ValueError(
            f"'andor' argument specified as {andor}. Expected one of {', '.join(andors)}"
        )

    ## Get database for specified species
    # Species shortcuts
    if species == "human":
        species = "homo_sapiens"
    if species == "mouse":
        species = "mus_musculus"

    # If a specific database is passed with the "/" at the end, remove it
    if "/" in species:
        species = species.split("/")[0]

    # In case species was passed with upper case letters
    species = species.lower()

    # Fetch all available databases
    databases = gget_species_options()
    db = []
    for datab in databases:
        if species in datab:
            db.append(datab)

    # Unless an unambigious mouse database is specified,
    # the standard core database will be used
    if len(db) > 1 and "mus_musculus" in species:
        db = f"mus_musculus_core_{ens_rel}_39"

    # Check for ambigious species matches in species other than mouse
    elif len(db) > 1 and "mus_musculus" not in species:
        raise ValueError(
            "Species matches more than one database.\n"
            "Please double-check spelling or pass specific CORE database.\n"
            "All available databases can be found here:\n"
            f"http://ftp.ensembl.org/pub/release-{ens_rel}/mysql/"
        )
    # Raise error if no matching database was found
    elif len(db) == 0:
        raise ValueError(
            "Species not found in database.\n"
            "Please double-check spelling or pass specific CORE database.\n"
            "All available databases can be found here:\n"
            f"http://ftp.ensembl.org/pub/release-{ens_rel}/mysql/"
        )

    else:
        db = db[0]

    logging.info(f"Fetching results from database: {db}")

    ## Connect to data base
    try:
        db_connection = sql.connect(
            host="ensembldb.ensembl.org", database=db, user="anonymous", password=""
        )
    except:
        try:
            # Try different port
            db_connection = sql.connect(
                host="ensembldb.ensembl.org",
                database=db,
                user="anonymous",
                password="",
                port=5306,
            )
        except Exception as e:
            logging.error(f"The Ensembl server returned the following error: {e}")

    ## Clean up list of searchwords
    # If single searchword passed as string, convert to list
    if type(searchwords) == str:
        searchwords = [searchwords]

    ## Find genes
    for i, searchword in enumerate(searchwords):
        if id_type == "gene":
            query = f"""
            SELECT gene.stable_id, xref.display_label, gene.description, xref.description, gene.biotype
            FROM gene
            LEFT JOIN xref ON gene.display_xref_id = xref.xref_id
            WHERE (gene.description LIKE '%{searchword}%' OR xref.description LIKE '%{searchword}%' OR xref.display_label LIKE '%{searchword}%')
            """

            # Fetch the search results from the host using the specified query
            df_temp = pd.read_sql(query, con=db_connection)
            # Order by ENSEMBL ID (I am using pandas for this instead of SQL to increase speed)
            df_temp = df_temp.sort_values("stable_id").reset_index(drop=True)

            # If andor="or", keep all results
            if andor == "or":
                # In the first iteration, make the search results equal to the master data frame
                if i == 0:
                    df = df_temp.copy()
                # Add new search results to master data frame
                else:
                    df = pd.concat([df, df_temp])

            # If andor="and", only keep overlap between results
            if andor == "and":
                # In the first iteration, make the search results equal to the master data frame
                if i == 0:
                    df = df_temp.copy()
                # Only keep overlapping results in master data frame
                else:
                    val = np.intersect1d(df["stable_id"], df_temp["stable_id"])
                    df = df[df.stable_id.isin(val)]

        if id_type == "transcript":
            query = f"""
            SELECT transcript.stable_id, xref.display_label, transcript.description, xref.description, transcript.biotype
            FROM transcript
            LEFT JOIN xref ON transcript.display_xref_id = xref.xref_id
            WHERE (transcript.description LIKE '%{searchword}%' OR xref.description LIKE '%{searchword}%' OR xref.display_label LIKE '%{searchword}%')
            """

            # Fetch the search results from the host using the specified query
            df_temp = pd.read_sql(query, con=db_connection)
            # Order by ENSEMBL ID (I am using pandas for this instead of SQL to increase speed)
            df_temp = df_temp.sort_values("stable_id").reset_index(drop=True)

            # If andor="or", keep all results
            if andor == "or":
                # In the first iteration, make the search results equal to the master data frame
                if i == 0:
                    df = df_temp.copy()
                # Add new search results to master data frame
                else:
                    df = pd.concat([df, df_temp])

            # If andor="and", only keep overlap between results
            if andor == "and":
                # In the first iteration, make the search results equal to the master data frame
                if i == 0:
                    df = df_temp.copy()
                # Only keep overlapping results in master data frame
                else:
                    val = np.intersect1d(df["stable_id"], df_temp["stable_id"])
                    df = df[df.stable_id.isin(val)]

    # Rename columns
    df = df.rename(
        columns={
            "stable_id": "ensembl_id",
            "display_label": "gene_name",
            "biotype": "biotype",
        }
    )
    # Changing description columns name by column index since they were returned with the same name ("description")
    df.columns.values[2] = "ensembl_description"
    df.columns.values[3] = "ext_ref_description"

    # Remove any duplicate search results from the master data frame and reset the index
    df = df.drop_duplicates().reset_index(drop=True)

    # If limit is not None, keep only the first {limit} rows
    if limit != None:
        # Print number of genes/transcripts found versus fetched
        logging.info(f"Returning {limit} matches of {len(df)} total matches found.")
        # Remove all but limit rows
        df = df.head(limit)

    else:
        # Print number of genes/transcripts fetched
        logging.info(f"Total matches found: {len(df)}.")

    # Print query time
    logging.info(f"Query time: {round(time.time() - start_time, 2)} seconds.")

    # Add URL to gene summary on Ensembl
    df["url"] = (
        "https://uswest.ensembl.org/"
        + "_".join(db.split("_")[:2])
        + "/Gene/Summary?g="
        + df["ensembl_id"]
    )

    if wrap_text:
        df_wrapped = df.copy()
        wrap_cols_func(
            df_wrapped, ["ensembl_description", "ext_ref_description", "url"]
        )

    if json:
        results_dict = json_package.loads(df.to_json(orient="records"))
        if save:
            with open("gget_search_results.json", "w", encoding="utf-8") as f:
                json_package.dump(results_dict, f, ensure_ascii=False, indent=4)

        # Return results in json format
        return results_dict

    else:
        # Save
        if save:
            df.to_csv("gget_search_results.csv", index=False)

        # Return data frame
        return df
